/*-
 * Copyright (c) 2000, 2001 Mitsuru IWASAKI <iwasaki@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id$
 */

#include <sys/param.h>
#include <sys/queue.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <assert.h>
#include <ctype.h>
#include <err.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <acpi.h>
#include <acnamesp.h>
#include <acdebug.h>

/*
 * Dummy DSDT Table Header
 */

ACPI_TABLE_HEADER	dummy_dsdt_table = {
	"DSDT", 123, 1, 123, "OEMID", "OEMTBLID", 1, "CRID", 1
};

/*
 * Region space I/O routines on virtual machine
 */

int	aml_debug_prompt = 1;

struct ACPIRegionContent {
	TAILQ_ENTRY(ACPIRegionContent) links;
	int		regtype;
	u_int32_t	addr;
	u_int8_t	value;
};

TAILQ_HEAD(ACPIRegionContentList, ACPIRegionContent);
struct	ACPIRegionContentList RegionContentList;

static int	 aml_simulation_initialized = 0;

static void	 aml_simulation_init(void);
static int	 aml_simulate_regcontent_add(int regtype, u_int32_t addr,
					     u_int8_t value);
static int	 aml_simulate_regcontent_read(int regtype, u_int32_t addr,
					      u_int8_t *valuep); 
static int	 aml_simulate_regcontent_write(int regtype, u_int32_t addr,
					       u_int8_t *valuep);
static u_int32_t aml_simulate_prompt(char *msg, u_int32_t def_val);
static void	 aml_simulation_regload(const char *dumpfile);
static void	 aml_simulation_regdump(const char *dumpfile);

static void
aml_simulation_init(void)
{

	aml_simulation_initialized = 1;
	TAILQ_INIT(&RegionContentList);
	aml_simulation_regload("region.ini");
}

static int
aml_simulate_regcontent_add(int regtype, u_int32_t addr, u_int8_t value)
{
	struct	ACPIRegionContent *rc;

	rc = malloc(sizeof(struct ACPIRegionContent));
	if (rc == NULL) {
		return (-1);	/* malloc fail */
	}
	rc->regtype = regtype;
	rc->addr = addr;
	rc->value = value;

	TAILQ_INSERT_TAIL(&RegionContentList, rc, links);
	return (0);
}

static int
aml_simulate_regcontent_read(int regtype, u_int32_t addr, u_int8_t *valuep)
{
	struct	ACPIRegionContent *rc;

	if (!aml_simulation_initialized) {
		aml_simulation_init();
	}
	TAILQ_FOREACH(rc, &RegionContentList, links) {
		if (rc->regtype == regtype && rc->addr == addr) {
			*valuep = rc->value;
			return (1);	/* found */
		}
	}

	*valuep = 0;
	return (aml_simulate_regcontent_add(regtype, addr, *valuep));
}

static int
aml_simulate_regcontent_write(int regtype, u_int32_t addr, u_int8_t *valuep)
{
	struct	ACPIRegionContent *rc;

	if (!aml_simulation_initialized) {
		aml_simulation_init();
	}
	TAILQ_FOREACH(rc, &RegionContentList, links) {
		if (rc->regtype == regtype && rc->addr == addr) {
			rc->value = *valuep;
			return (1);	/* exists */
		}
	}

	return (aml_simulate_regcontent_add(regtype, addr, *valuep));
}

static u_int32_t
aml_simulate_prompt(char *msg, u_int32_t def_val)
{
	char		buf[16], *ep;
	u_int32_t	val;

	val = def_val;
	printf("DEBUG");
	if (msg != NULL) {
		printf("%s", msg);
	}
	printf("(default: 0x%x / %u) >>", val, val);
	fflush(stdout);

	bzero(buf, sizeof buf);
	while (1) {
		if (read(0, buf, sizeof buf) == 0) {
			continue;
		}
		if (buf[0] == '\n') {
			break;	/* use default value */
		}
		if (buf[0] == '0' && buf[1] == 'x') {
			val = strtoq(buf, &ep, 16);
		} else {
			val = strtoq(buf, &ep, 10);
		}
		break;
	}
	return (val);
}

static void
aml_simulation_regload(const char *dumpfile)
{
	char	buf[256], *np, *ep;
	struct	ACPIRegionContent rc;
	FILE	*fp;

	if (!aml_simulation_initialized) {
		return;
	}

	if ((fp = fopen(dumpfile, "r")) == NULL) {
		return;
	}

	while (fgets(buf, sizeof buf, fp) != NULL) {
		np = buf;
		/* reading region type */
		rc.regtype = strtoq(np, &ep, 10);
		if (np == ep) {
			continue;
		}
		np = ep;

		/* reading address */
		rc.addr = strtoq(np, &ep, 16);
		if (np == ep) {
			continue;
		}
		np = ep;

		/* reading value */
		rc.value = strtoq(np, &ep, 16);
		if (np == ep) {
			continue;
		}
		aml_simulate_regcontent_write(rc.regtype, rc.addr, &rc.value);
	}

	fclose(fp);
}

static void
aml_simulation_regdump(const char *dumpfile)
{
	struct	ACPIRegionContent *rc;
	FILE	*fp;

	if (!aml_simulation_initialized) {
		return;
	}
	if ((fp = fopen(dumpfile, "w")) == NULL) {
		warn("%s", dumpfile);
		return;
	}
	while (!TAILQ_EMPTY(&RegionContentList)) {
		rc = TAILQ_FIRST(&RegionContentList);
		fprintf(fp, "%d	0x%x	0x%x\n",
		    rc->regtype, rc->addr, rc->value);
		TAILQ_REMOVE(&RegionContentList, rc, links);
		free(rc);
	}

	fclose(fp);
	TAILQ_INIT(&RegionContentList);
}

/*
 * Space handlers on virtual machine
 */

static ACPI_STATUS	 VMSpaceHandler(UINT32 SpaceID, UINT32 Function,
					ACPI_PHYSICAL_ADDRESS Address,
					UINT32 BitWidth, UINT32 *Value,
					int Prompt);
static ACPI_STATUS	 VMSystemMemorySpaceHandler(UINT32 Function,
					ACPI_PHYSICAL_ADDRESS Address,
					UINT32 BitWidth, UINT32 *Value,
					void *HandlerContext,
					void *RegionContext);
static ACPI_STATUS	 VMSystemIOSpaceHandler(UINT32 Function,
					ACPI_PHYSICAL_ADDRESS Address,
					UINT32 BitWidth, UINT32 *Value,
					void *HandlerContext,
					void *RegionContext);
static ACPI_STATUS	 VMPciConfigSpaceHandler(UINT32 Function,
					ACPI_PHYSICAL_ADDRESS Address,
					UINT32 BitWidth, UINT32 *Value,
					void *HandlerContext,
					void *RegionContext);

static ACPI_STATUS
VMSpaceHandler(
	UINT32			SpaceID,
	UINT32			Function,
	ACPI_PHYSICAL_ADDRESS	Address,
	UINT32			BitWidth,
	UINT32			*Value,
	int			Prompt)
{
	int		state;
	u_int8_t	val;
	u_int32_t	value, i;
	char		msg[256];

	switch (Function) {
	case ADDRESS_SPACE_READ:
		value = 0;
		for (i = 0; (i * 8) < BitWidth; i++) {
			state = aml_simulate_regcontent_read(SpaceID,
							     Address + i, &val);
			if (state == -1) {
				return (AE_NO_MEMORY);
			}
			value |= val << (i * 8);
		}
		*Value = value;
		if (Prompt) {
			sprintf(msg, "[read (%d, %2d, 0x%x)]",
				SpaceID, BitWidth, Address);
			*Value = aml_simulate_prompt(msg, value);
			if (*Value != value) {
				return(VMSpaceHandler(SpaceID, ADDRESS_SPACE_WRITE,
						      Address, BitWidth, Value, 0));
			}
		}
		break;

	case ADDRESS_SPACE_WRITE:
		value = *Value;
		if (Prompt) {
			sprintf(msg, "[write(%d, %2d, 0x%x)]",
				SpaceID, BitWidth, Address);
			value = aml_simulate_prompt(msg, *Value);
		}
		for (i = 0; (i * 8) < BitWidth; i++) {
			val = value & 0xff;
			state = aml_simulate_regcontent_write(SpaceID,
							      Address + i, &val);
			if (state == -1) {
				return (AE_NO_MEMORY);
			}
			value = value >> 8;
		}
	}

	return (AE_OK);
}

static ACPI_STATUS
VMSystemMemorySpaceHandler(
	UINT32			Function,
	ACPI_PHYSICAL_ADDRESS	Address,
	UINT32			BitWidth,
	UINT32			*Value,
	void			*HandlerContext,
	void			*RegionContext)
{
	return(VMSpaceHandler(ADDRESS_SPACE_SYSTEM_MEMORY, Function, Address, BitWidth, Value, aml_debug_prompt));
}

static ACPI_STATUS
VMSystemIOSpaceHandler(
	UINT32			Function,
	ACPI_PHYSICAL_ADDRESS	Address,
	UINT32			BitWidth,
	UINT32			*Value,
	void			*HandlerContext,
	void			*RegionContext)
{
	return(VMSpaceHandler(ADDRESS_SPACE_SYSTEM_IO, Function, Address, BitWidth, Value, aml_debug_prompt));
}

static ACPI_STATUS
VMPciConfigSpaceHandler(
	UINT32			Function,
	ACPI_PHYSICAL_ADDRESS	Address,
	UINT32			BitWidth,
	UINT32			*Value,
	void			*HandlerContext,
	void			*RegionContext)
{
	return(VMSpaceHandler(ADDRESS_SPACE_PCI_CONFIG, Function, Address, BitWidth, Value, aml_debug_prompt));
}

/*
 * Load DSDT data file and invoke debugger
 */

static int
load_dsdt(const char *dsdtfile)
{
	u_int8_t		*code, *amlptr;
	struct stat		 sb;
	int			 fd;
	int			 error;
	ACPI_TABLE_HEADER	*tableptr;

	fd = open(dsdtfile, O_RDONLY, 0);
	if (fd == -1) {
		perror("open");
		return (-1);
	}
	if (fstat(fd, &sb) == -1) {
		perror("fstat");
		return (-1);
	}
	code = mmap(NULL, (size_t)sb.st_size, PROT_READ, MAP_PRIVATE, fd, (off_t)0);
	if (code == NULL) {
		perror("mmap");
		return (-1);
	}
	if ((error = AcpiInitializeSubsystem()) != AE_OK) {
		return (-1);
	}

	/*
	 * make sure DSDT data contains table header or not.
	 */
	if (strncmp((char *)code, "DSDT", 4) == 0) {
		tableptr = (ACPI_TABLE_HEADER *)code;
		amlptr = code + sizeof(ACPI_TABLE_HEADER);
	} else {
		tableptr = malloc(sizeof(ACPI_TABLE_HEADER) + sb.st_size);
		if (tableptr == NULL) {
			return (-1);
		}

		amlptr = (u_int8_t *)tableptr;
		amlptr += sizeof(ACPI_TABLE_HEADER);
		dummy_dsdt_table.Length = sizeof(ACPI_TABLE_HEADER) + sb.st_size;
		bcopy(&dummy_dsdt_table, tableptr, sizeof(ACPI_TABLE_HEADER));
		bcopy(code, amlptr, sb.st_size);
	}


	AcpiGbl_AcpiTables[ACPI_TABLE_DSDT].Pointer = AcpiGbl_DSDT = tableptr;
	AcpiGbl_AcpiTables[ACPI_TABLE_DSDT].AmlPointer = amlptr;
	AcpiGbl_AcpiTables[ACPI_TABLE_DSDT].AmlLength = tableptr->Length - sizeof(ACPI_TABLE_HEADER);

	/*
	 * Install the virtual machine version of address space handlers.
	 */
	if ((error = AcpiInstallAddressSpaceHandler(ACPI_ROOT_OBJECT,
						    ADDRESS_SPACE_SYSTEM_MEMORY,
						    VMSystemMemorySpaceHandler,
						    NULL, NULL)) != AE_OK) {
		fprintf(stderr, "could not initialise SystemMemory handler: %d\n", error);
		return (-1);
	}
	if ((error = AcpiInstallAddressSpaceHandler(ACPI_ROOT_OBJECT,
						    ADDRESS_SPACE_SYSTEM_IO,
						    VMSystemIOSpaceHandler,
						    NULL, NULL)) != AE_OK) {
		fprintf(stderr, "could not initialise SystemIO handler: %d\n", error);
		return (-1);
	}
	if ((error = AcpiInstallAddressSpaceHandler(ACPI_ROOT_OBJECT,
						    ADDRESS_SPACE_PCI_CONFIG,
						    VMPciConfigSpaceHandler,
						    NULL, NULL)) != AE_OK) {
		fprintf(stderr, "could not initialise PciConfig handler: %d\n", error);
		return (-1);
	}

	AcpiNsLoadNamespace();
	AcpiDbInitialize();
	AcpiGbl_DebuggerConfiguration = 0;
	AcpiDbUserCommands(NULL, NULL);

	return (0);
}

static void
usage(const char *progname)
{

	printf("usage: %s dsdt_file\n", progname);
	exit(1);
}

int
main(int argc, char *argv[])
{
	int	status;
	char	*progname;

	progname = argv[0];

	if (argc == 1) {
		usage(progname);
	}

	AcpiDbgLevel = DEBUG_DEFAULT;
	status = load_dsdt(argv[1]);
	if (status == 0) {
		aml_simulation_regdump("region.dmp");
	}

	return (status);
}
